@@ -20,16 +20,13 @@ import re
 import sys
 
 from dulwich import diff_tree
+from dulwich import index as d_index
 from dulwich import objects
+from dulwich import porcelain
 from dulwich import repo
 
 LOG = logging.getLogger(__name__)
 
-# What does a pre-release version number look like?
-PRE_RELEASE_RE = re.compile('''
-    \.(\d+(?:[ab]|rc)+\d*)$
-''', flags=re.VERBOSE | re.UNICODE)
-
 
 def _parse_version(v):
     parts = v.split('.') + ['0', '0', '0']
@@ -209,6 +206,150 @@ def _aggregate_changes(walk_entry, changes, notesdir):
     return results
 
 
+class _ChangeTracker(object):
+
+    def __init__(self):
+        # Track the versions we have seen and the earliest version for
+        # which we have seen a given note's unique id.
+        self.versions = []
+        self.earliest_seen = collections.OrderedDict()
+        # Remember the most current filename for each id, to allow for
+        # renames.
+        self.last_name_by_id = {}
+        # Remember uniqueids that have had files deleted.
+        self.uniqueids_deleted = set()
+
+    def _common(self, uniqueid, sha, version):
+        if version not in self.versions:
+            self.versions.append(version)
+        # Update the "earliest" version where a UID appears
+        # every time we see it, because we are scanning the
+        # history in reverse order so "early" items come
+        # later.
+        if uniqueid in self.earliest_seen:
+            LOG.debug('%s: resetting earliest reference from %s to %s for %s',
+                      uniqueid, self.earliest_seen[uniqueid], version, sha)
+        else:
+            LOG.debug('%s: setting earliest reference to %s for %s',
+                      uniqueid, version, sha)
+        self.earliest_seen[uniqueid] = version
+
+    def add(self, filename, sha, version):
+        uniqueid = _get_unique_id(filename)
+        self._common(uniqueid, sha, version)
+        LOG.info('%s: adding %s from %s',
+                 uniqueid, filename, version)
+
+        # If we have recorded that a UID was deleted, that
+        # means that was the last change made to the file and
+        # we can ignore it.
+        if uniqueid in self.uniqueids_deleted:
+            LOG.debug(
+                '%s: has already been deleted, ignoring this change',
+                uniqueid,
+            )
+            return
+
+        # A note is being added in this commit. If we have
+        # not seen it before, it was added here and never
+        # changed.
+        if uniqueid not in self.last_name_by_id:
+            self.last_name_by_id[uniqueid] = (filename, sha)
+            LOG.info(
+                '%s: new %s in commit %s',
+                uniqueid, filename, sha,
+            )
+        else:
+            LOG.debug(
+                '%s: add for file we have already seen',
+                uniqueid,
+            )
+
+    def rename(self, filename, sha, version):
+        uniqueid = _get_unique_id(filename)
+        self._common(uniqueid, sha, version)
+
+        # If we have recorded that a UID was deleted, that
+        # means that was the last change made to the file and
+        # we can ignore it.
+        if uniqueid in self.uniqueids_deleted:
+            LOG.debug(
+                '%s: has already been deleted, ignoring this change',
+                uniqueid,
+            )
+            return
+
+        # The file is being renamed. We may have seen it
+        # before, if there were subsequent modifications,
+        # so only store the name information if it is not
+        # there already.
+        if uniqueid not in self.last_name_by_id:
+            self.last_name_by_id[uniqueid] = (filename, sha)
+            LOG.info(
+                '%s: update to %s in commit %s',
+                uniqueid, filename, sha,
+            )
+        else:
+            LOG.debug(
+                '%s: renamed file already known with the new name',
+                uniqueid,
+            )
+
+    def modify(self, filename, sha, version):
+        uniqueid = _get_unique_id(filename)
+        self._common(uniqueid, sha, version)
+
+        # If we have recorded that a UID was deleted, that
+        # means that was the last change made to the file and
+        # we can ignore it.
+        if uniqueid in self.uniqueids_deleted:
+            LOG.debug(
+                '%s: has already been deleted, ignoring this change',
+                uniqueid,
+            )
+            return
+
+        # An existing file is being modified. We may have
+        # seen it before, if there were subsequent
+        # modifications, so only store the name
+        # information if it is not there already.
+        if uniqueid not in self.last_name_by_id:
+            self.last_name_by_id[uniqueid] = (filename, sha)
+            LOG.info(
+                '%s: update to %s in commit %s',
+                uniqueid, filename, sha,
+            )
+        else:
+            LOG.debug(
+                '%s: modified file already known',
+                uniqueid,
+            )
+
+    def delete(self, filename, sha, version):
+        uniqueid = _get_unique_id(filename)
+        self._common(uniqueid, sha, version)
+        # This file is being deleted without a rename. If
+        # we have already seen the UID before, that means
+        # that after the file was deleted another file
+        # with the same UID was added back. In that case
+        # we do not want to treat it as deleted.
+        #
+        # Never store deleted files in last_name_by_id so
+        # we can safely use all of those entries to build
+        # the history data.
+        if uniqueid not in self.last_name_by_id:
+            self.uniqueids_deleted.add(uniqueid)
+            LOG.info(
+                '%s: note deleted in %s',
+                uniqueid, sha,
+            )
+        else:
+            LOG.debug(
+                '%s: delete for file re-added after the delete',
+                uniqueid,
+            )
+
+
 class RenoRepo(repo.Repo):
 
     # Populated by _load_tags().
@@ -283,7 +424,23 @@ class RenoRepo(repo.Repo):
             return tree
 
     def get_file_at_commit(self, filename, sha):
-        "Return the contents of the file if it exists at the commit, or None."
+        """Return the contents of the file.
+
+        If sha is None, return the working copy of the file. If the
+        file cannot be read from the working dir, return None.
+
+        If the sha is not None and the file exists at the commit,
+        return the data from the stored blob. If the file does not
+        exist at the commit, return None.
+
+        """
+        if sha is None:
+            # Get the copy from the working directory.
+            try:
+                with open(os.path.join(self.path, filename), 'r') as f:
+                    return f.read()
+            except IOError:
+                return None
         # Get the tree associated with the commit identified by the
         # input SHA, then look through the items in the tree to find
         # the one with the path matching the filename. Take the
@@ -309,6 +466,18 @@ class Scanner(object):
         self.conf = conf
         self.reporoot = self.conf.reporoot
         self._repo = RenoRepo(self.reporoot)
+        self.release_tag_re = re.compile(
+            self.conf.release_tag_re,
+            flags=re.VERBOSE | re.UNICODE,
+        )
+        self.pre_release_tag_re = re.compile(
+            self.conf.pre_release_tag_re,
+            flags=re.VERBOSE | re.UNICODE,
+        )
+        self.branch_name_re = re.compile(
+            self.conf.branch_name_re,
+            flags=re.VERBOSE | re.UNICODE,
+        )
 
     def _get_ref(self, name):
         if name:
@@ -338,6 +507,10 @@ class Scanner(object):
         branch_head = self._get_ref(branch)
         return self._repo.get_walker(branch_head)
 
+    def _get_valid_tags_on_commit(self, sha):
+        return [tag for tag in self._repo.get_tags_on_commit(sha)
+                if self.release_tag_re.match(tag)]
+
     def _get_tags_on_branch(self, branch):
         "Return a list of tag names on the given branch."
         results = []
@@ -345,7 +518,7 @@ class Scanner(object):
             # shas_to_tags has encoded versions of the shas
             # but the commit object gives us a decoded version
             sha = c.commit.sha().hexdigest().encode('ascii')
-            tags = self._repo.get_tags_on_commit(sha)
+            tags = self._get_valid_tags_on_commit(sha)
             results.extend(tags)
         return results
 
@@ -361,7 +534,7 @@ class Scanner(object):
             # shas_to_tags has encoded versions of the shas
             # but the commit object gives us a decoded version
             sha = commit.sha().hexdigest().encode('ascii')
-            tags = self._repo.get_tags_on_commit(sha)
+            tags = self._get_valid_tags_on_commit(sha)
             if tags:
                 if count:
                     val = '{}-{}'.format(tags[-1], count)
@@ -376,6 +549,27 @@ class Scanner(object):
                 commit = None
         return '0.0.0'
 
+    def _strip_pre_release(self, tag):
+        """Return tag with pre-release identifier removed if present."""
+        pre_release_match = self.pre_release_tag_re.search(tag)
+        if pre_release_match:
+            try:
+                start = pre_release_match.start('pre_release')
+                end = pre_release_match.end('pre_release')
+            except IndexError:
+                raise ValueError(
+                    ("The pre-release tag regular expression, {!r}, is missing"
+                     " a group named 'pre_release'.").format(
+                        self.pre_release_tag_re.pattern
+                    )
+                )
+            else:
+                stripped_tag = tag[:start] + tag[end:]
+        else:
+            stripped_tag = tag
+
+        return stripped_tag
+
     def _get_branch_base(self, branch):
         "Return the tag at base of the branch."
         # Based on
@@ -394,7 +588,7 @@ class Scanner(object):
             if c.commit.sha().hexdigest() in master_commits:
                 # We got to this commit via the branch, but it is also
                 # on master, so this is the base.
-                tags = self._repo.get_tags_on_commit(
+                tags = self._get_valid_tags_on_commit(
                     c.commit.sha().hexdigest().encode('ascii'))
                 if tags:
                     return tags[-1]
@@ -508,9 +702,47 @@ class Scanner(object):
         "Return true if the file exists at the given commit."
         return bool(self.get_file_at_commit(filename, sha))
 
-    @staticmethod
-    def _find_scan_stop_point(earliest_version, versions_by_date,
-                              collapse_pre_releases):
+    def _get_series_branches(self):
+        "Get branches matching the branch_name_re config option."
+        refs = self._repo.get_refs()
+        LOG.debug('refs %s', list(refs.keys()))
+        branch_names = set()
+        for r in refs.keys():
+            name = None
+            r = r.decode('utf-8')
+            if r.startswith('refs/remotes/origin/'):
+                name = r[20:]
+            elif r.startswith('refs/heads/'):
+                name = r[11:]
+            if name and self.branch_name_re.search(name):
+                branch_names.add(name)
+        return list(sorted(branch_names))
+
+    def _get_earlier_branch(self, branch):
+        "Return the name of the branch created before the given branch."
+        # FIXME(dhellmann): Assumes branches come in order based on
+        # name. That may not be true for projects that branch based on
+        # version numbers instead of names.
+        if branch.startswith('origin/'):
+            branch = branch[7:]
+        LOG.debug('looking for the branch before %s', branch)
+        branch_names = self._get_series_branches()
+        if branch not in branch_names:
+            LOG.debug('Could not find branch %r among %s',
+                      branch, branch_names)
+            return None
+        LOG.debug('found branches %s', branch_names)
+        current = branch_names.index(branch)
+        if current == 0:
+            # This is the first branch.
+            LOG.debug('%s appears to be the first branch', branch)
+            return None
+        previous = branch_names[current - 1]
+        LOG.debug('found earlier branch %s', previous)
+        return previous
+
+    def _find_scan_stop_point(self, earliest_version, versions_by_date,
+                              collapse_pre_releases, branch):
         """Return the version to use to stop the scan.
 
         Use the list of versions_by_date to get the tag with a
@@ -525,6 +757,7 @@ class Scanner(object):
         :param collapse_pre_releases: Boolean indicating whether we are
             collapsing pre-releases or not. If false, the next tag
             is used, regardless of its version.
+        :param branch: The name of the branch we are scanning.
 
         """
         if not earliest_version:
@@ -536,7 +769,17 @@ class Scanner(object):
             # The version we were given is not present, use a full
             # scan.
             return None
-        if not collapse_pre_releases:
+        # We need to look for the previous branch's root.
+        if branch and branch != 'master':
+            previous_branch = self._get_earlier_branch(branch)
+            if not previous_branch:
+                # This was the first branch, so scan the whole
+                # history.
+                return None
+            previous_base = self._get_branch_base(previous_branch)
+            return previous_base
+        is_pre_release = bool(self.pre_release_tag_re.search(earliest_version))
+        if is_pre_release and not collapse_pre_releases:
             # We just take the next tag.
             return versions_by_date[idx]
         # We need to look for a different version.
@@ -566,9 +809,15 @@ class Scanner(object):
         collapse_pre_releases = self.conf.collapse_pre_releases
         stop_at_branch_base = self.conf.stop_at_branch_base
 
-        LOG.info('scanning %s/%s (branch=%s)',
+        LOG.info('scanning %s/%s (branch=%s earliest_version=%s)',
                  reporoot.rstrip('/'), notesdir.lstrip('/'),
-                 branch or '*current*')
+                 branch or '*current*', earliest_version)
+
+        # Determine the current version, which might be an unreleased or
+        # dev version if there are unreleased commits at the head of the
+        # branch in question.
+        current_version = self._get_current_version(branch)
+        LOG.debug('current repository version: %s' % current_version)
 
         # Determine all of the tags known on the branch, in their date
         # order. We scan the commit history in topological order to ensure
@@ -584,27 +833,43 @@ class Scanner(object):
         # If the user has told us where to stop, use that as the
         # default.
         scan_stop_tag = self._find_scan_stop_point(
-            earliest_version, versions_by_date, collapse_pre_releases)
+            earliest_version, versions_by_date,
+            collapse_pre_releases, branch)
 
         # If the user has not told us where to stop, try to work it
-        # out for ourselves. If branch is set and is not "master",
-        # then we want to stop at the version before the tag at the
-        # base of the branch, which involves a bit of searching.
-        if (stop_at_branch_base and
-                (not earliest_version) and branch and (branch != 'master')):
+        # out for ourselves.
+        if not branch and not earliest_version and stop_at_branch_base:
+            # On the current branch, stop at the point where the most
+            # recent branch was created, if we can find one.
+            LOG.debug('working on current branch without earliest_version')
+            branches = self._get_series_branches()
+            if branches:
+                LOG.debug('looking at base of %s to stop scanning master',
+                          branches[-1])
+                scan_stop_tag = self._get_branch_base(branches[-1])
+                earliest_version = current_version
+        elif branch and stop_at_branch_base and not earliest_version:
+            # If branch is set and is not "master",
+            # then we want to stop at the version before the tag at the
+            # base of the branch, which involves a bit of searching.
             LOG.debug('determining earliest_version from branch')
-            earliest_version = self._get_branch_base(branch)
+            branch_base = self._get_branch_base(branch)
             scan_stop_tag = self._find_scan_stop_point(
-                earliest_version, versions_by_date,
-                collapse_pre_releases)
+                branch_base, versions_by_date,
+                collapse_pre_releases, branch)
+            if not scan_stop_tag:
+                earliest_version = branch_base
+            else:
+                idx = versions_by_date.index(scan_stop_tag)
+                earliest_version = versions_by_date[idx - 1]
             if earliest_version and collapse_pre_releases:
-                if PRE_RELEASE_RE.search(earliest_version):
+                if self.pre_release_tag_re.search(earliest_version):
                     # The earliest version won't actually be the pre-release
                     # that might have been tagged when the branch was created,
                     # but the final version. Strip the pre-release portion of
                     # the version number.
-                    earliest_version = '.'.join(
-                        earliest_version.split('.')[:-1]
+                    earliest_version = self._strip_pre_release(
+                        earliest_version
                     )
         if earliest_version:
             LOG.info('earliest version to include is %s', earliest_version)
@@ -613,31 +878,61 @@ class Scanner(object):
         if scan_stop_tag:
             LOG.info('stopping scan at %s', scan_stop_tag)
 
-        versions = []
-        earliest_seen = collections.OrderedDict()
-
-        # Determine the current version, which might be an unreleased or
-        # dev version if there are unreleased commits at the head of the
-        # branch in question. Since the version may not already be known,
-        # make sure it is in the list of versions by date. And since it is
-        # the most recent version, go ahead and insert it at the front of
-        # the list.
-        current_version = self._get_current_version(branch)
-        LOG.debug('current repository version: %s' % current_version)
+        # Since the version may not already be known, make sure it is
+        # in the list of versions by date. And since it is the most
+        # recent version, go ahead and insert it at the front of the
+        # list.
         if current_version not in versions_by_date:
             versions_by_date.insert(0, current_version)
+        versions_by_date.insert(0, '*working-copy*')
 
-        # Remember the most current filename for each id, to allow for
-        # renames.
-        last_name_by_id = {}
+        # Track the versions we have seen and the earliest version for
+        # which we have seen a given note's unique id.
+        tracker = _ChangeTracker()
 
-        # Remember uniqueids that have had files deleted.
-        uniqueids_deleted = set()
+        # Process the local index, if we are scanning the current
+        # branch.
+        if not branch:
+            prefix = notesdir.rstrip('/') + '/'
+            index = self._repo.open_index()
 
+            # Pretend anything known to the repo and changed but not
+            # staged is part of the fake version '*working-copy*'.
+            LOG.debug('scanning unstaged changes')
+            for fname in d_index.get_unstaged_changes(index, self.reporoot):
+                fname = fname.decode('utf-8')
+                LOG.debug('found unstaged file %s', fname)
+                if fname.startswith(prefix) and _note_file(fname):
+                    fullpath = os.path.join(self.reporoot, fname)
+                    if os.path.exists(fullpath):
+                        LOG.debug('found file %s', fullpath)
+                        tracker.add(fname, None, '*working-copy*')
+                    else:
+                        LOG.debug('deleted file %s', fullpath)
+                        tracker.delete(fname, None, '*working-copy*')
+
+            # Pretend anything in the index is part of the fake
+            # version "*working-copy*".
+            LOG.debug('scanning staged schanges')
+            changes = porcelain.get_tree_changes(self._repo)
+            for fname in changes['add']:
+                fname = fname.decode('utf-8')
+                if fname.startswith(prefix) and _note_file(fname):
+                    tracker.add(fname, None, '*working-copy*')
+            for fname in changes['modify']:
+                fname = fname.decode('utf-8')
+                if fname.startswith(prefix) and _note_file(fname):
+                    tracker.modify(fname, None, '*working-copy*')
+            for fname in changes['delete']:
+                fname = fname.decode('utf-8')
+                if fname.startswith(prefix) and _note_file(fname):
+                    tracker.delete(fname, None, '*working-copy*')
+
+        # Process the git commit history.
         for counter, entry in enumerate(self._topo_traversal(branch), 1):
 
             sha = entry.commit.id
-            tags_on_commit = self._repo.get_tags_on_commit(sha)
+            tags_on_commit = self._get_valid_tags_on_commit(sha)
 
             LOG.debug('%06d %s %s', counter, sha, tags_on_commit)
 
@@ -651,116 +946,35 @@ class Scanner(object):
                 LOG.info('%06d %s updating current version to %s',
                          counter, sha, current_version)
 
-            # Remember each version we have seen.
-            if current_version not in versions:
-                LOG.debug('%s is a new version' % current_version)
-                versions.append(current_version)
-
-            # Look for changes to notes files in this commit.
+            # Look for changes to notes files in this commit. The
+            # change has only the basename of the path file, so we
+            # need to prefix that with the notesdir before giving it
+            # to the tracker.
             changes = _changes_in_subdir(self._repo, entry, notesdir)
             for change in _aggregate_changes(entry, changes, notesdir):
                 uniqueid = change[0]
 
-                # Update the "earliest" version where a UID appears
-                # every time we see it, because we are scanning the
-                # history in reverse order so "early" items come
-                # later.
-                LOG.debug('%s: setting earliest reference to %s',
-                          uniqueid, current_version)
-                earliest_seen[uniqueid] = current_version
-
                 c_type = change[1]
 
-                # If we have recorded that a UID was deleted, that
-                # means that was the last change made to the file and
-                # we can ignore it.
-                if uniqueid in uniqueids_deleted:
-                    LOG.debug(
-                        '%s: has already been deleted, ignoring this change',
-                        uniqueid,
-                    )
-                    continue
-
                 if c_type == diff_tree.CHANGE_ADD:
-                    # A note is being added in this commit. If we have
-                    # not seen it before, it was added here and never
-                    # changed.
-                    if uniqueid not in last_name_by_id:
-                        path, sha = change[-2:]
-                        fullpath = os.path.join(notesdir, path)
-                        last_name_by_id[uniqueid] = (fullpath,
-                                                     sha.decode('ascii'))
-                        LOG.info(
-                            '%s: update to %s in commit %s',
-                            uniqueid, path, sha,
-                        )
-                    else:
-                        LOG.debug(
-                            '%s: add for file we have already seen',
-                            uniqueid,
-                        )
+                    path, blob_sha = change[-2:]
+                    fullpath = os.path.join(notesdir, path)
+                    tracker.add(fullpath, sha, current_version)
 
                 elif c_type == diff_tree.CHANGE_DELETE:
-                    # This file is being deleted without a rename. If
-                    # we have already seen the UID before, that means
-                    # that after the file was deleted another file
-                    # with the same UID was added back. In that case
-                    # we do not want to treat it as deleted.
-                    #
-                    # Never store deleted files in last_name_by_id so
-                    # we can safely use all of those entries to build
-                    # the history data.
-                    if uniqueid not in last_name_by_id:
-                        uniqueids_deleted.add(uniqueid)
-                        LOG.info(
-                            '%s: note deleted in %s',
-                            uniqueid, sha,
-                        )
-                    else:
-                        LOG.debug(
-                            '%s: delete for file re-added after the delete',
-                            uniqueid,
-                        )
+                    path = change[-1]
+                    fullpath = os.path.join(notesdir, path)
+                    tracker.delete(fullpath, sha, current_version)
 
                 elif c_type == diff_tree.CHANGE_RENAME:
-                    # The file is being renamed. We may have seen it
-                    # before, if there were subsequent modifications,
-                    # so only store the name information if it is not
-                    # there already.
-                    if uniqueid not in last_name_by_id:
-                        path, sha = change[-2:]
-                        fullpath = os.path.join(notesdir, path)
-                        last_name_by_id[uniqueid] = (fullpath,
-                                                     sha.decode('ascii'))
-                        LOG.info(
-                            '%s: update to %s in commit %s',
-                            uniqueid, path, sha,
-                        )
-                    else:
-                        LOG.debug(
-                            '%s: renamed file already known with the new name',
-                            uniqueid,
-                        )
+                    path, blob_sha = change[-2:]
+                    fullpath = os.path.join(notesdir, path)
+                    tracker.rename(fullpath, sha, current_version)
 
                 elif c_type == diff_tree.CHANGE_MODIFY:
-                    # An existing file is being modified. We may have
-                    # seen it before, if there were subsequent
-                    # modifications, so only store the name
-                    # information if it is not there already.
-                    if uniqueid not in last_name_by_id:
-                        path, sha = change[-2:]
-                        fullpath = os.path.join(notesdir, path)
-                        last_name_by_id[uniqueid] = (fullpath,
-                                                     sha.decode('ascii'))
-                        LOG.info(
-                            '%s: update to %s in commit %s',
-                            uniqueid, path, sha,
-                        )
-                    else:
-                        LOG.debug(
-                            '%s: modified file already known',
-                            uniqueid,
-                        )
+                    path, blob_sha = change[-2:]
+                    fullpath = os.path.join(notesdir, path)
+                    tracker.modify(fullpath, sha, current_version)
 
                 else:
                     raise ValueError(
@@ -777,13 +991,15 @@ class Scanner(object):
         # Invert earliest_seen to make a list of notes files for each
         # version.
         files_and_tags = collections.OrderedDict()
-        for v in versions:
+        for v in tracker.versions:
             files_and_tags[v] = []
         # Produce a list of the actual files present in the repository. If
         # a note is removed, this step should let us ignore it.
-        for uniqueid, version in earliest_seen.items():
+        for uniqueid, version in tracker.earliest_seen.items():
             try:
-                base, sha = last_name_by_id[uniqueid]
+                base, sha = tracker.last_name_by_id[uniqueid]
+                LOG.debug('%s: sorting %s into version %s',
+                          uniqueid, base, version)
                 files_and_tags[version].append((base, sha))
             except KeyError:
                 # Unable to find the file again, skip it to avoid breaking
@@ -796,6 +1012,7 @@ class Scanner(object):
         # Combine pre-releases into the final release, if we are told to
         # and the final release exists.
         if collapse_pre_releases:
+            LOG.debug('collapsing pre-release versions into final releases')
             collapsing = files_and_tags
             files_and_tags = collections.OrderedDict()
             for ov in versions_by_date:
@@ -803,13 +1020,12 @@ class Scanner(object):
                     # We don't need to collapse this one because there are
                     # no notes attached to it.
                     continue
-                pre_release_match = PRE_RELEASE_RE.search(ov)
+                pre_release_match = self.pre_release_tag_re.search(ov)
                 LOG.debug('checking %r', ov)
                 if pre_release_match:
                     # Remove the trailing pre-release part of the version
                     # from the string.
-                    pre_rel_str = pre_release_match.groups()[0]
-                    canonical_ver = ov[:-len(pre_rel_str)].rstrip('.')
+                    canonical_ver = self._strip_pre_release(ov)
                     if canonical_ver not in versions_by_date:
                         # This canonical version was never tagged, so we
                         # do not want to collapse the pre-releases. Reset
@@ -823,12 +1039,16 @@ class Scanner(object):
                     files_and_tags[canonical_ver] = []
                 files_and_tags[canonical_ver].extend(collapsing[ov])
 
+        LOG.debug('files_and_tags %s',
+                  {k: len(v) for k, v in files_and_tags.items()})
         # Only return the parts of files_and_tags that actually have
         # filenames associated with the versions.
+        LOG.debug('trimming')
         trimmed = collections.OrderedDict()
         for ov in versions_by_date:
             if not files_and_tags.get(ov):
                 continue
+            LOG.debug('keeping %s', ov)
             # Sort the notes associated with the version so they are in a
             # deterministic order, to avoid having the same data result in
             # different output depending on random factors. Earlier
@@ -842,6 +1062,7 @@ class Scanner(object):
             # If we have been told to stop at a version, we can do that
             # now.
             if earliest_version and ov == earliest_version:
+                LOG.debug('stopping trimming at %s', earliest_version)
                 break
 
         LOG.debug(
